from functools import partial

from ruamel.yaml import YAML
from ruamel.yaml.dumper import SafeDumper
from ruamel.yaml.nodes import ScalarNode
from .aws import get_parameter
from .aws.session import create_session
from .aws.pstore import create_kms_client
from .aws.pstore import create_ssm_client


class SsmValue:
    def __init__(
            self, key, profile=None, region=None, role_arn=None,
            session=None, ssm_client=None):
        self.key = key
        self.resolved = False
        self.m_value = None
        self.profile = profile
        self.region = region
        self.role_arn = role_arn
        self.session = session
        self.ssm_client = ssm_client

    @property
    def value(self):
        if self.resolved:
            return self.m_value
        self.m_value = get_parameter(
            self.key, profile=self.profile, region=self.region,
            session=self.session, client=self.ssm_client)
        self.resolved = True
        return self.m_value


def ssm_scalar_constructor(
        loader, node, profile=None, region=None, role_arn=None,
        session=None, ssm_client=None, special_values=None):
    key = loader.construct_scalar(node)
    result = SsmValue(key, profile, region, role_arn, session, ssm_client)
    if special_values is not None:
        special_values.append(result)
    return result


def ssm_representer(dumper: SafeDumper, value: SsmValue) -> ScalarNode:
    """
    round-trip serialize an ssm value
    """
    return dumper.represent_scalar("!ssm", value.key)


class KmsWrappedSecret:
    def __init__(
            self, encrypted_value, profile=None, region=None,
            role_arn=None, kms_key=None, session=None, kms_client=None):
        self.encrypted_value = encrypted_value
        self.kms_key = kms_key
        self.resolved = False
        self.m_value = None
        self.profile = profile
        self.region = region
        self.role_arn = role_arn
        self.session = session
        self.kms_client = kms_client

    @property
    def value(self):
        from murmuration import kms_wrapped
        if self.resolved:
            return self.m_value
        self.m_value = kms_wrapped.decrypt(
            self.encrypted_value, self.region, self.profile,
            self.session, client=self.kms_client)
        self.resolved = True
        return self.m_value


def kms_wrapped_scalar_constructor(
        loader, node, profile=None, region=None, role_arn=None,
        kms_key=None, session=None, kms_client=None, special_values=None):
    encrypted_value = loader.construct_scalar(node)
    result = KmsWrappedSecret(
        encrypted_value,
        profile=profile,
        region=region,
        role_arn=role_arn,
        kms_key=kms_key,
        session=session, kms_client=kms_client)
    if special_values is not None:
        special_values.append(result)
    return result


def kms_wrapped_representer(
        dumper: SafeDumper,
        value: KmsWrappedSecret) -> ScalarNode:
    """
    round-trip serialize a kms-wrapped secret
    """
    return dumper.represent_scalar("!kms_wrapped", value.encrypted_value)


class MasterKeyedSecret:
    def __init__(
            self, encrypted_value, master_key=None):
        self.encrypted_value = encrypted_value
        self.master_key = master_key
        self.resolved = False
        self.m_value = None

    @property
    def value(self):
        from murmuration import gcm
        if self.resolved:
            return self.m_value
        self.m_value = gcm.decrypt(self.encrypted_value, self.master_key.value)
        self.resolved = True
        return self.m_value


def master_keyed_scalar_constructor(
        loader, node, master_key=None, special_values=None):
    encrypted_value = loader.construct_scalar(node)
    result = MasterKeyedSecret(
        encrypted_value,
        master_key)
    if special_values is not None:
        special_values.append(result)
    return result


def master_keyed_representer(
        dumper: SafeDumper,
        value: MasterKeyedSecret) -> ScalarNode:
    """
    round-trip serialize a master-keyed secret
    """
    return dumper.represent_scalar("!secret", value.encrypted_value)


def reset_special_value_clients(
        special_values, profile, region, role_arn, master_key):
    session = create_session(
        profile=profile, region=region, role_arn=role_arn)
    kms_client = create_kms_client(session=session)
    ssm_client = create_ssm_client(session=session)
    for x in special_values:
        if isinstance(x, SsmValue):
            x.ssm_client = ssm_client
        elif isinstance(x, KmsWrappedSecret):
            x.kms_client = kms_client
        elif isinstance(x, MasterKeyedSecret):
            x.master_key = master_key


class Yaml(YAML):
    def __init__(self, profile=None, region=None, role_arn=None,
                 kms_key=None, master_key=None, session=None,
                 ssm_client=None, kms_client=None, width=4096):
        super().__init__(typ='rt')
        self.explicit_start = True
        self.preserve_quotes = True
        self.width = width
        self.indent(sequence=4, mapping=2, offset=2)
        self.special_values = []
        ssm_fn = partial(
            ssm_scalar_constructor,
            profile=profile,
            region=region,
            role_arn=role_arn,
            session=session,
            ssm_client=ssm_client,
            special_values=self.special_values)
        kms_wrapped_fn = partial(
            kms_wrapped_scalar_constructor,
            profile=profile,
            region=region,
            role_arn=role_arn,
            kms_key=kms_key,
            session=session,
            kms_client=kms_client,
            special_values=self.special_values)
        master_keyed_wrapped_fn = partial(
            master_keyed_scalar_constructor,
            master_key=master_key, special_values=self.special_values)
        self.constructor.add_constructor('!secret', master_keyed_wrapped_fn)
        self.constructor.add_constructor('!ssm', ssm_fn)
        self.constructor.add_constructor('!kms_wrapped', kms_wrapped_fn)
        self.representer.add_representer(SsmValue, ssm_representer)
        self.representer.add_representer(
            KmsWrappedSecret, kms_wrapped_representer)
        self.representer.add_representer(
            MasterKeyedSecret,
            master_keyed_representer)


def dump_yaml(x, filename):
    y = Yaml()
    if isinstance(filename, str):
        with open(filename, 'w') as f:
            y.dump(x, f)
    else:
        y.dump(x, filename)


def basic_loader():
    from yaml import SafeLoader

    class WaddleLoader(SafeLoader):
        special_values = []

    ssm_fn = partial(
        ssm_scalar_constructor,
        special_values=WaddleLoader.special_values)
    kms_wrapped_fn = partial(
        kms_wrapped_scalar_constructor,
        special_values=WaddleLoader.special_values)
    master_keyed_wrapped_fn = partial(
        master_keyed_scalar_constructor,
        special_values=WaddleLoader.special_values)
    WaddleLoader.add_constructor('!secret', master_keyed_wrapped_fn)
    WaddleLoader.add_constructor('!ssm', ssm_fn)
    WaddleLoader.add_constructor('!kms_wrapped', kms_wrapped_fn)

    return WaddleLoader


def load_yaml_basic(filename):
    from yaml import load
    WaddleLoader = basic_loader()
    with open(filename, 'r') as f:
        data = load(f, Loader=WaddleLoader)
    master_key = data.get('meta', {}).get('master_key')
    special_values = WaddleLoader.special_values
    if master_key:
        for x in special_values:
            if isinstance(x, MasterKeyedSecret):
                x.master_key = master_key
    return data, master_key, special_values



from boto3.session import Session


def create_session(
        region=None, profile=None, *,
        role_arn=None, session=None):
    """
    A handy helper function that will create the
    boto session using our waddle-level settings
    """
    from ..settings import get_region, get_profile, get_role_arn
    if not session:
        profile = profile or get_profile()
        region = region or get_region()
        role_arn = role_arn or get_role_arn()
        session = Session(
            region_name=region,
            profile_name=profile,
        )
        if role_arn:
            sts = session.client('sts')
            response = sts.assume_role(
                RoleArn=role_arn,
                RoleSessionName='waddle')
            creds = response['Credentials']
            session = Session(
                aws_access_key_id=creds['AccessKeyId'],
                aws_secret_access_key=creds['SecretAccessKey'],
                aws_session_token=creds['SessionToken'],
            )
    return session



from collections.abc import Mapping
from itertools import chain
import logging
import os
import re
from typing import Dict, Any, List, Optional, Tuple
from ruamel.yaml.comments import CommentedMap

from .yaml_base import KmsWrappedSecret
from .yaml_base import SsmValue
from .yaml_base import MasterKeyedSecret


__all__ = [
    'Bunch',
    'BunchList',
]
log = logging.getLogger(__name__)
AnyDict = Dict[str, Any]
dict_class = CommentedMap


def wrap(val, obj_wrapper=None):
    if isinstance(val, Mapping):
        return Bunch(val) if obj_wrapper is None else obj_wrapper(val)
    if isinstance(val, list):
        return BunchList(val)
    if isinstance(val, (SsmValue, KmsWrappedSecret, MasterKeyedSecret)):
        return val.value
    return val


class BunchList:
    def __init__(self, values, obj_wrapper=None):
        "make iterables into lists"
        if not isinstance(values, list):
            values = list(values)
        self.values = values
        self._obj_wrapper = obj_wrapper

    def __repr__(self):
        return repr(self.values)

    def __eq__(self, other):
        if isinstance(other, BunchList):
            return other.values == self.values
        # make sure we still equal to a dict with the same data
        return other == self.values

    def __ne__(self, other):
        return not self == other

    def __getitem__(self, key):
        value = self.values[key]
        if isinstance(key, slice):
            return BunchList(value, obj_wrapper=self._obj_wrapper)
        return wrap(value, self._obj_wrapper)

    def __setitem__(self, k, value):
        self.values[k] = value

    def __iter__(self):
        return map(lambda i: wrap(i, self._obj_wrapper), self.values)

    def __len__(self):
        return len(self.values)

    def __nonzero__(self):
        return bool(self.values)
    __bool__ = __nonzero__

    def __getattr__(self, name):
        return getattr(self.values, name)

    def __getstate__(self):
        return self.values, self._obj_wrapper

    def __setstate__(self, state):
        self.values, self._obj_wrapper = state


class Bunch:
    """
    Bunch pattern that allows us to reference settings values
    file in a pathy manner.  e.g.,

        x = Bunch()
        x.path.to.key = value
        x.path.greeting = 'hello'
        x['path.other_greeting'] = 'hello world'
    """
    def _set(self, key, value):
        """
        This is a useful function if you want to set
        an attribute on the underlying object so that
        the value isn't stored in the `self.values` dictionary.
        We need this because we override `__setattr__` below.
        """
        super().__setattr__(key, value)

    def _get(self, key):
        """
        This is a useful function if you want to get
        an attribute on the underlying object.
        We need this because we override `__getattr__` below.
        """
        return self.__dict__.get(key)

    def __init__(self, values=None):
        if values is None:
            self.values: AnyDict = dict_class()
        else:
            self.values: AnyDict = values

    def __contains__(self, key):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            return pieces[-1] in parent
        return False

    def __nonzero__(self):
        return bool(self.values)

    __bool__ = __nonzero__

    def __dir__(self):
        # introspection for auto-complete in IPython etc
        return chain(self.values, super().__dir__())

    def __eq__(self, other):
        if isinstance(other, Bunch):
            return other.values == self.values
        # make sure we still equal to a dict with the same data
        return other == self.values

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        r = repr(self.values).replace('ordereddict', '')
        if len(r) > 60:
            r = r[:60] + '...])'
        return r

    def __getstate__(self) -> Tuple[Any]:
        # n.b., trailing comma is intentional
        # since pickle needs a tuple return value
        return self.values,  # pylint: disable=R1707

    def __setstate__(self, state):
        self.values = state[0]

    def __getattr__(self, key):
        if self._is_reserved(key):
            return self._get(key)
        try:
            return wrap(self.values[key])
        except KeyError:
            self.values[key] = dict_class()
            return wrap(self.values[key])

    def __delattr__(self, key):
        try:
            del self.values[key]
        except KeyError as ex:
            raise AttributeError(
                f'{self.__class__.__name__} object has no attribute {key}'
            ) from ex

    def __getitem__(self, key):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            return wrap(parent.get(pieces[-1]))
        return None

    def __setitem__(self, key, value):
        pieces = key.split('.')
        parent = self.create_to_parent(pieces)
        parent[pieces[-1]] = value

    def __delitem__(self, key):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            del parent[pieces[-1]]

    def _is_reserved(self, key):
        if key in [ 'values' ]:
            return True
        return key in self.__dict__ and not key.startswith('__')

    def __setattr__(self, key, value):
        if self._is_reserved(key):
            self._set(key, value)
            return
        self.values[key] = value

    def create_to_parent(self, pieces: List[str]) -> AnyDict:
        parent: AnyDict = self.values
        for y in pieces[:-1]:
            if y not in parent:
                child = parent[y] = dict_class()
            else:
                child = parent[y]
            parent = child
        return parent

    def walk_to_parent(self, pieces: List[str]) -> Optional[AnyDict]:
        parent: Dict[str, Any] = self.values
        for x in pieces[:-1]:
            parent = parent.get(x)
            if not isinstance(parent, dict):
                return None
        return parent

    def get(self, key, default=None):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            return wrap(parent.get(pieces[-1], default))
        return wrap(default)

    def setdefault(self, key, value):
        pieces = key.split('.')
        parent = self.create_to_parent(pieces)
        key = pieces[-1]
        return parent.setdefault(key, value)

    def env(self):
        splitter = re.compile(r'[_.]')

        def fn(key, default=None):
            value = os.environ.get(key)
            if not value:
                pieces = splitter.split(key.lower())
                parent: AnyDict = self.walk_to_parent(pieces)
                if not isinstance(parent, dict):
                    return None
                value = parent.get(pieces[-1], default)
            return wrap(value)
        return fn

    def keys(self):
        return self.values.keys()

    def __iter__(self):
        return iter(self.values)

    def items(self, values=None, prefix=None):
        values = values or self.values
        prefix = prefix or []
        for key, value in values.items():
            # the shape key appears when debugging in jetbrains ides
            if key == 'shape' and not prefix:
                continue  # pragma: no cover
            if isinstance(value, Mapping):
                yield from self.items(
                    values=value,
                    prefix=prefix + [ key ])
            else:
                yield '.'.join(prefix + [ key ]), value


    @classmethod
    def unwrap_value(cls, value):
        if isinstance(value, (SsmValue, KmsWrappedSecret, MasterKeyedSecret)):
            return value.value
        return value

    @classmethod
    def serialize_list(cls, values):
        result = []
        for value in values:
            if isinstance(value, Mapping):
                x = cls.serialize_dict(value)
            elif isinstance(value, List):
                x = cls.serialize_list(value)
            else:
                x = cls.unwrap_value(value)
            result.append(x)
        return result

    @classmethod
    def serialize_dict(cls, values):
        result = {}
        for key, value in values.items():
            if isinstance(value, Mapping):
                result[key] = cls.serialize_dict(values=value)
            elif isinstance(value, List):
                result[key] = cls.serialize_list(value)
            else:
                result[key] = cls.unwrap_value(value)
        return result


    def to_dict(self, values=None, result=None):
        values = values or self.values
        result = {}
        for key, value in values.items():
            # the shape key appears when debugging in jetbrains ides
            if key in ('shape', 'meta'):
                continue  # pragma: no cover
            if isinstance(value, Mapping):
                result[key] = self.serialize_dict(value)
            elif isinstance(value, List):
                result[key] = self.serialize_list(value)
            else:
                result[key] = self.unwrap_value(value)
        return result



import os
import random
import sys


__all__ = [
    'load_config',
    'start_notification',
    'end_notification',
    'long_interactive',
    'generate_master_key',
]


def load_config(*filenames, decrypt=True):
    """
    Loads the specified config files if they exist in the order
    specified.  This is useful for having "cascading" configs
    and also for suppressing warnings from `ParamBunch.from_file`
    """
    from .param_bunch import ParamBunch
    conf = ParamBunch()
    for x in filenames:
        if os.path.exists(x):
            conf.from_file(x, decrypt=decrypt)
    return conf


def start_notification(action, key, encrypted=False):
    from halo import Halo
    encrypted = ' (encrypted)' if encrypted else ''
    if sys.stdout.isatty():  # pragma: no cover
        spinner = Halo(f'{action} {key}{encrypted}', spinner='dots')
        spinner.start()
        return spinner
    message = f'{action} {key}{encrypted} ....'
    print(message, end='')
    return None


SUCCESS = '\N{heavy check mark}\N{vs16}'
FAILURE = '\N{ballot x}\N{vs16}'


def end_notification(spinner, success=True):
    from halo.halo import colored_frame
    mark = SUCCESS if success else FAILURE
    if sys.stdout.isatty():  # pragma: no-cover
        if success:
            spinner.succeed()
        else:
            mark = colored_frame(mark, 'red')
            spinner.stop_and_persist(mark)
    else:
        print(mark, end='')


class long_interactive:
    def __init__(self, action, key, encrypted=False):
        self.spinners = []
        self.action = action
        self.key = key
        self.encrypted = encrypted
        self.success = None

    def succeeded(self):
        self.success = True

    def failed(self):
        self.success = False

    def __enter__(self):
        spinner = start_notification(self.action, self.key, self.encrypted)
        self.spinners.append(spinner)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        spinner = self.spinners.pop()
        end_notification(spinner, self.success)


def generate_master_key(length=32):
    chars = (
        '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        '_+-/;,.=~!@#$%^&*'
    )
    rg = [ random.choice(chars) for _ in range(length) ]
    return ''.join(rg)



import logging.config
import re
import sys
import click
from murmuration import kms_wrapped
from murmuration import gcm

from .settings import setup_logging
from .param_bunch import ParamBunch
from .yaml_base import KmsWrappedSecret, MasterKeyedSecret
from .utils import long_interactive
from .utils import generate_master_key
from . import version


@click.group(name='waddle')
@click.version_option(version)
def main():
    "cli for managing waddle config files"


def is_secret(key):
    matcher = re.compile('.*(key|secret|token|password).*')
    return matcher.match(key)


@main.command(name='add-secret')
@click.argument('key', metavar='db.password')
@click.option('-f', '--filename', metavar='/path/to/config_file.yml',
              type=click.Path(exists=True), required=True)
def add_secret(filename, key):
    """
    Adds an encrypted secret to the specified configuration file

    Example:
        waddle add-secret -f conf/dev.yml db.password
    """
    setup_logging()
    tty = sys.stdin.isatty()
    log = logging.getLogger(__name__)
    x = ParamBunch()
    x.from_file(
        filename=filename, decrypt=False, handle_tags=False,
        round_trip=True)
    kms_key = x.get('meta.kms_key')
    master_key = None
    try:
        _, _, master_key = x.original_value('meta.master_key')
    except KeyError:
        pass
    if not kms_key and not master_key:
        log.warning('%s needs one of kms_key or master_key.', filename)
        return
    if tty:  # pragma: no cover
        print(f'Enter value for [{key}]: ', end='', file=sys.stderr)
        sys.stderr.flush()
    # stdin = os.fdopen(sys.stdin.fileno(), 'rb', 0)
    plain_text = sys.stdin.readline().rstrip()
    # plain_text = plain_text.decode('utf-8').rstrip()

    region = x.get('meta.region')
    profile = x.get('meta.profile')
    role_arn = x.get('meta.role_arn')
    if master_key:
        if isinstance(master_key, KmsWrappedSecret):
            master_key = master_key.value
        encrypted_value = gcm.encrypt(plain_text, master_key, 'waddle')
        x[key] = MasterKeyedSecret(encrypted_value)
    else:
        encrypted_value = kms_wrapped.encrypt(
            plain_text, kms_key,
            region, profile, role_arn=role_arn)
        x[key] = KmsWrappedSecret(encrypted_value)
    log.info('saving')
    x.save(filename)
    log.info('done')


@main.command(name='encrypt')
@click.option('-f', '--filename', metavar='/path/to/config_file.yml',
              type=click.Path(exists=True), required=True)
def encrypt(filename):
    """
    Encrypts values for any key that that has the following keywords in it:
      * key
      * password
      * token
      * secret

    Example:
        waddle encrypt -f conf/dev.yml
    """
    setup_logging()
    log = logging.getLogger(__name__)
    x = ParamBunch()
    x.load(filename=filename, decrypt=True, round_trip=True)
    kms_key = x.get('meta.kms_key')
    master_key = None
    try:
        _, _, master_key = x.original_value('meta.master_key')
    except KeyError:
        pass
    if not kms_key and not master_key:
        log.warning('%s needs one of kms_key or master_key.', filename)
        return
    region = x.get('meta.region')
    profile = x.get('meta.profile')
    role_arn = x.get('meta.role_arn')
    values = []
    for key, value in x.items():
        values.append([key, value])
    for key, value in values:
        if is_secret(key) and not key.startswith('meta.'):
            if master_key:
                if isinstance(master_key, KmsWrappedSecret):
                    master_key = master_key.value
                encrypted_value = gcm.encrypt(value, master_key, 'waddle')
                x[key] = MasterKeyedSecret(encrypted_value)
            else:
                encrypted_value = kms_wrapped.encrypt(
                    value, kms_key, region=region, profile=profile,
                    role_arn=role_arn)
                x[key] = KmsWrappedSecret(encrypted_value)
    x.save(filename)


@main.command(name='deploy')
@click.option('-f', '--filename', metavar='/path/to/config_file.yml',
              type=click.Path(exists=True), required=True)
@click.option('-e', '--encrypted', is_flag=True)
def deploy(filename, encrypted):
    """
    Deploys a locally stored config file to aws:

    Example:
        waddle deploy -f conf/dev.yml

    Use the --encrypt flag to push all values as kms encrypted to
    parameter store.
    """
    x = ParamBunch(filename=filename)
    x.to_aws(force_encryption=encrypted)


@main.command(name='undeploy')
@click.option('-f', '--filename', metavar='/path/to/config_file.yml',
              type=click.Path(exists=True), required=True)
def undeploy(filename):
    """
    Deploys a locally stored config file to aws:

    Example:
        waddle deploy -f conf/dev.yml
    """
    x = ParamBunch(filename=filename)
    x.delete_from_aws()


@main.command(name='add_master_key')
@click.option('-f', '--filename', metavar='/path/to/config_file.yml',
              type=click.Path(exists=True), required=True)
@click.option('-r', '--reencrypt', is_flag=True, show_default=True,
              default=False, help='will reencrypt all secrets with the added '
                                  'master key')
def add_master_key(filename, reencrypt):
    """
    adds a master key to the specified config file.
    """
    setup_logging()
    log = logging.getLogger(__name__)
    x = ParamBunch(filename=filename, round_trip=True)
    kms_key = x.get('meta.kms_key')
    if not kms_key:
        log.warning('%s needs a kms_key.', filename)
        return
    try:
        _, _, master_key = x.original_value('meta.master_key')
    except KeyError:
        master_key = None
    if master_key:
        log.info('rotating existing master key')
        reencrypt = True
    region = x.get('meta.region')
    profile = x.get('meta.profile')
    role_arn = x.get('meta.role_arn')
    new_master_key = generate_master_key()
    if reencrypt:
        for key in x.encrypted:
            with long_interactive('encrypting', key=key) as ctx:
                value = x.get(key)
                encrypted_value = gcm.encrypt(value, new_master_key, 'waddle')
                x[key] = MasterKeyedSecret(encrypted_value)
                ctx.succeeded()
    with long_interactive('saving', key=filename) as ctx:
        encrypted_value = kms_wrapped.encrypt(
            new_master_key, alias=kms_key, profile=profile, region=region,
            role_arn=role_arn)
        x.meta.master_key = KmsWrappedSecret(encrypted_value)
        x.save(filename)
        ctx.succeeded()


if __name__ == "__main__":
    main()



from collections.abc import Mapping
import logging
import os
import re

from ruamel.yaml.comments import CommentedMap
from murmuration import kms_wrapped
from .bunch import Bunch
from .aws import yield_parameters
from .aws import put_parameter
from .aws import delete_parameters
from .aws import ssm_key
from .aws import create_ssm_client
from .yaml_base import Yaml, dump_yaml
from .yaml_base import SsmValue, KmsWrappedSecret, MasterKeyedSecret
from .yaml_base import load_yaml_basic, reset_special_value_clients


__all__ = [
    'ParamBunch',
]


dict_class = CommentedMap


class ParamBunch(Bunch):
    def __init__(
            self, values=None, prefix=None,
            filename=None, round_trip=False):
        super().__init__(values)
        super()._set('original_values', values)
        super()._set('encrypted', [])
        super()._set('spinner', None)
        if prefix or filename:
            self.load(prefix=prefix, filename=filename, round_trip=round_trip)

    def _is_reserved(self, key):
        if key in [ 'values', 'original_value', 'encrypted', 'spinner', ]:
            return True
        return key in self.__dict__ and not key.startswith('__')

    def aws_items(self, values=None, prefix=None):
        prefix = prefix or [ '', self.meta.namespace ]
        for key, value in self.items(values, prefix):
            if '.meta.' in key:
                continue
            key = key.replace('.', '/')
            yield key, value

    def file_items(self, values=None, prefix=None):
        meta_prefix = re.compile(r'^\.?meta\.')
        for key, value in self.items(values, prefix):
            if meta_prefix.match(key):
                continue
            yield key, value
        yield from self.items(values=self.meta.values, prefix=[ 'meta' ])

    def to_dict(self, values=None, result=None, **kwargs):
        result = super().to_dict(values=values, result=result, **kwargs)
        return result

    @staticmethod
    def _traverse(d, prefix=None):
        prefix = prefix or []
        for key, value in d.items():
            if isinstance(value, Mapping):
                yield from ParamBunch._traverse(value, prefix + [ key ])
            else:
                yield '.'.join(prefix + [ key ]), value

    @staticmethod
    def try_decrypt(
            value, region=None, profile=None,
            session=None, client=None):
        # an artifact of murmuration encryption is that all
        # encrypted values will start with 'AQI', so fast fail
        # out of decrypt
        if isinstance(value, str) and value[:3] == 'AQI':
            try:
                value = kms_wrapped.decrypt(
                    value, region, profile, session, client=client)
            except ValueError:
                pass
        return value

    @classmethod
    def get_master_key(cls, meta, profile, region, role_arn, kms_key):
        master_key = meta.get('master_key')
        if isinstance(master_key, KmsWrappedSecret):
            master_key.profile = profile
            master_key.region = region
            master_key.role_arn = role_arn
            master_key.kms_key = kms_key
            return master_key
        return master_key

    def from_file(  # pylint: disable=unused-argument
            self, filename,
            decrypt=True,
            round_trip=False, **kwargs):
        # late load logger in case anyone sets up logging
        # later
        log = logging.getLogger(__name__)
        if not os.path.exists(filename):
            log.warning('waddle: could not find %s', filename)
            return

        special_values, master_key = [], None
        if round_trip:
            with open(filename, 'r', encoding='utf-8') as f:
                y = Yaml()
                data = y.load(f)
        else:
            data, master_key, special_values = load_yaml_basic(filename)
        meta = data.get('meta', {})
        region = meta.get('region')
        profile = meta.get('profile')
        kms_key = meta.get('kms_key')
        role_arn = meta.get('role_arn')
        if round_trip:
            master_key = self.get_master_key(
                meta, profile, region, role_arn, kms_key)
            special_values = y.special_values
        reset_special_value_clients(
            special_values, profile=profile, region=region,
            role_arn=role_arn, master_key=master_key)
        super()._set('original_values', data)
        values = []
        for key, value in ParamBunch._traverse(data):
            if self._is_reserved(key):
                raise KeyError(f'`{key}` is not a valid key name')
            if key.startswith('meta.'):
                self[key] = value
            else:
                values.append((key, value))
        self.handle_file_values(values, decrypt)

    def handle_file_values(self, values, decrypt, session=None, client=None):
        from murmuration.aws import kms_client
        region = self.get('meta.region')
        profile = self.get('meta.profile')
        client = kms_client(
            region=region, profile=profile,
            session=session, client=client)
        for key, value in values:
            if not decrypt:
                self[key] = value
                continue
            new_value = self[key] = ParamBunch.try_decrypt(
                value, region, profile,
                session=session, client=client)
            tagged_classes = (SsmValue, KmsWrappedSecret, MasterKeyedSecret)
            tagged_value = isinstance(value, tagged_classes)
            if new_value != value or tagged_value:
                self.encrypted.append(key)

    def load(self, prefix=None, filename=None, decrypt=True, round_trip=False):
        if prefix:
            self.from_aws(prefix)
        if filename:
            self.from_file(filename, decrypt, round_trip)

    def from_aws(self, prefix):
        if not prefix.startswith('/'):
            prefix = f'/{prefix}'
        prefix = prefix.replace('.', '/')
        for key, value, type_ in yield_parameters(prefix):
            if type_ == 'SecureString':
                self.encrypted.append(key)
            elif type_ == 'StringList':
                value = value.split(',')
            self[key] = value
        self.meta.namespace = prefix[1:].replace('/', '.')

    def _encrypted_keys(self):
        namespace = self.get('meta.namespace', '')
        ms_encrypted = set()
        for x in self.encrypted:
            ms_encrypted.add(ssm_key(namespace, x))
        return ms_encrypted

    def to_aws(
            self, force_encryption=False, verbose=True,
            profile=None, region=None,
            role_arn=None, session=None, ssm_client=None):
        ssm_client = create_ssm_client(
            profile=profile, region=region, role_arn=role_arn,
            session=session, client=ssm_client)
        ms_encrypted = self._encrypted_keys()
        kms_key = self.get('meta.kms_key')
        for key, value in self.aws_items():
            encrypted = key in ms_encrypted
            encrypted = encrypted or force_encryption
            put_parameter(
                key, value, kms_key, encrypted, verbose,
                session=session, client=ssm_client)

    def delete_from_aws(
            self, verbose=True,
            profile=None, region=None, role_arn=None,
            session=None, ssm_client=None):
        ssm_client = create_ssm_client(
            profile=profile, region=region, role_arn=role_arn,
            session=session, client=ssm_client)
        keys = [ key for key, _ in self.aws_items() ]
        delete_parameters(
            *keys, verbose=verbose,
            session=session, client=ssm_client)

    def original_value(self, key):
        data = self.original_values
        if key in data:
            return data, key, data[key]
        pieces = key.split('.')
        for x in pieces[:-1]:
            data = data[x]
        key = pieces[-1]
        return data, key, data[key]

    def original_parent(self, key):
        data = self.original_values
        pieces = key.split('.')
        try:
            for x in pieces[:-1]:
                data = data[x]
            return data, pieces[-1]
        except (KeyError, TypeError):
            return self.original_values, key

    def fill_back(self):
        updated_values = []
        new_values = []
        for key, value in self.items():
            try:
                parent, key, original_value = self.original_value(key)
                if value != original_value:
                    updated_values.append((parent, key, value))
            except (KeyError, TypeError):
                new_values.append((key, value))
        self.handle_updates(updated_values)
        self.handle_new(new_values)

    def handle_updates(self, updated_values):
        # pylint: disable=access-member-before-definition
        for parent, key, value in updated_values:
            parent[key] = value
        if self.original_values is None:
            self.original_values = dict_class()

    def handle_new(self, new_values):
        for key, value in new_values:
            parent, x = self.original_parent(key)
            parent[x] = value

    def save(self, filename):
        self.fill_back()
        dump_yaml(self.original_values, filename)



import logging


# settings to control aws connection
aws_profile = None
aws_secret_access_key = None
aws_access_key_id = None
aws_region = None
aws_role_arn = None
verbose_mode = False
test_mode = False


def set_profile(profile):
    global aws_profile
    aws_profile = profile


def get_profile():
    return aws_profile


def set_region(region):
    global aws_region
    aws_region = region


def get_region():
    return aws_region


def set_role_arn(role_arn):
    global aws_role_arn
    aws_role_arn = role_arn


def get_role_arn():
    return aws_role_arn


def set_verbose_mode():
    return verbose_mode


def get_verbose_mode():
    return verbose_mode


def setup_logging():
    if test_mode:
        return
    logging.config.dictConfig({
        'version': 1,
        'root': {
            'handlers': ['waddle'],
            'level': 'INFO',
        },
        'handlers': {
            'waddle': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',
                'formatter': 'waddle',
            }
        },
        'formatters': {
            'waddle': {
                'format': '[{asctime}.{msecs:03.0f}] [{name}] {message}',
                'datefmt': '%Y-%m-%d %H:%M:%S',
                'style': '{',
            },
        },
    })


def get_test_mode():
    return test_mode


def set_test_mode(boo):
    global test_mode
    test_mode = boo



from typing import Generator, Tuple
from murmuration.helpers import prefix_alias

try:
    from contextlib import nullcontext
except ImportError:
    class nullcontext:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass


__all__ = [
    'yield_parameters',
    'put_parameter',
    'get_parameter',
    'delete_parameters',
    'waddle_key',
    'ssm_key',
    'create_ssm_client',
    'create_kms_client',
]

StrTuple = Tuple[str, str, str]


def waddle_key(prefix, key):
    return key.replace(prefix, '').replace('/', '.')[1:]


def ssm_key(prefix, key):
    prefix = f'/{prefix}' if prefix else ''
    return f'{prefix}/{key}'.replace('.', '/')


def create_kms_client(
        profile=None, region=None, *,
        role_arn=None, session=None, client=None,):
    from .session import create_session
    if client:
        return client
    session = create_session(
        profile=profile,
        region=region,
        role_arn=role_arn,
        session=session)
    return session.client('kms')


def create_ssm_client(
        profile=None, region=None, *,
        role_arn=None, session=None, client=None):
    from .session import create_session
    if client:
        return client
    session = create_session(
        profile=profile,
        region=region,
        role_arn=role_arn,
        session=session)
    return session.client('ssm')


def get_parameter(key, profile=None, region=None, session=None, client=None):
    ssm = create_ssm_client(
        profile=profile, region=region, session=session, client=client)
    try:
        result = ssm.get_parameter(Name=key, WithDecryption=True)
        return result['Parameter']['Value']
    except:  # noqa, pylint: disable=bare-except
        return None


def put_parameter(
        key, value, kms_key, encrypted, verbose=False,
        region=None, profile=None, session=None, client=None):
    from ..utils import long_interactive
    ctx = nullcontext()
    if verbose:
        ctx = long_interactive('pushing', key, encrypted)
    with ctx:
        ssm = create_ssm_client(
            region=region, profile=profile,
            session=session, client=client)
        params = {}
        if isinstance(value, list):
            value = ','.join([ f'{x}' for x in value ])
            params['Type'] = 'StringList'
        elif encrypted:
            params['Type'] = 'SecureString'
            if kms_key:
                params['KeyId'] = prefix_alias(kms_key)
        else:
            params['Type'] = 'String'
        result = ssm.put_parameter(
            Name=key,
            Value=value,
            Overwrite=True,
            **params)
        if verbose:
            ctx.succeeded()
    return result


def delete_parameters(*keys, verbose=False, region=None, profile=None,
                      session=None, client=None):
    from ..utils import long_interactive
    ssm = create_ssm_client(
        region=region, profile=profile, session=session, client=client)
    n_start = 1
    while keys:
        rg = keys[:10]
        ctx = nullcontext()
        if verbose:
            message = f'{n_start}=>{n_start + len(rg) - 1}'
            ctx = long_interactive('deleting keys', message)
        with ctx:
            ssm.delete_parameters(Names=rg)
            keys = keys[10:]
            n_start += len(rg)
            if verbose:
                ctx.succeeded()


def yield_parameters(
        prefix, decrypt=True,
        profile=None, region=None, session=None,
        client=None) -> Generator[StrTuple, None, None]:
    ssm = create_ssm_client(
        profile=profile, region=region, session=session, client=client)
    paginator = ssm.get_paginator('get_parameters_by_path')
    for page in paginator.paginate(
            Path=prefix,
            Recursive=True,
            WithDecryption=decrypt,
            PaginationConfig={
                'PageSize': 10,
            }):
        for x in page['Parameters']:
            key = x['Name']
            value = x['Value']
            key = waddle_key(prefix, key)
            yield key, value, x['Type']



"""
Contains a number of useful modules for handling
configuration and encryption using aws parameter store
and kms
"""
from .bunch import *  # noqa: F401, F403
from .param_bunch import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
version = '1.8'



from setuptools import setup, find_packages

version = '1.8'

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='waddle',
    version=version,
    description="A pathy wrapper around aws parameter store",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Developers',
      'Topic :: Software Development :: Libraries :: Python Modules',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
      'License :: OSI Approved :: BSD License',
    ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords='aws python parameter-store kms',
    author='Preetam Shingavi',
    author_email='p.shingavi@yahoo.com',
    url='https://github.com/angry-penguins/waddle',
    license='BSD',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'pyyaml',
        'boto3>=1.9.0',
        'click>=7.0',
        'murmuration>=1.4',
        'ruamel.yaml>=0.15.87',
        'halo>=0.0.26',
    ],
    entry_points={
        'console_scripts': [
            'waddle=waddle.cli:main',
        ]
    })
